/*
    Trabalho Prtico III - IPDI
    -------------------------
    Anderson de Rezende Rocha <anderson.rocha@ic.unicamp.br>
    Anlise de texturas por histograma de somas e diferenas

    OBSERVAO
    ----------
    Imagem[i,j] = *(Imagem + i * nc + j)
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "imagem.h"
#include "morfo.h"
#include "utils.h"
#include "tratadorErros.h"

// Constante para alguma referncia de nmeros muito grandes
#define INFINITO 99999999
// Referncia ao ponteiro /dev/null
#define null 0

// Histograma de somas
#define HIST_SOMA 0
// Histograma de diferenas
#define HIST_DIF 1

typedef struct T_Prototipo
{
    // media, contraste, homogen, energia, variancia, correlacao
    // entropia
    double *descritor;
} Prototipo;

static int TAM_JANELA;
static int MAX_NIVEL_CINZA;
static int DIRX;
static int DIRY;
static int DESCRITOR;
static int NRO_PROTOTIPOS;
static int NRO_DESCRITORES;
static char ARQUIVO_PROTOTIPOS[40];
static int TODAS_DIRECOES;

void detectarTexturas(int *imagem, int *imagemSaida, int nl, int nc);
void criarHistogramaSD(int janela[][TAM_JANELA], int nl, int nc, int t1, int t2, int *hs, int *hd);
double avaliarDescritor(int *hs, int *hd, int descritor);
void configurarParametros(char *nomeArquivo);
void subQuantizarImagem(int *imagem, int nl, int nc,  int maxCinza);
void configurarPrototipos(char *nomeArquivo, Prototipo *p);
void classificar(int *imagem, int *imagemSaida, int nl, int nc, Prototipo *p);
double distanciaEuclidiana(Prototipo *origem, Prototipo *destino, int n);
void lerPrototipos(Prototipo *p);

// Funo principal
int main (int argc, char *argv[])
{
    int OK, nc, nl, mn, i, j;
    int *imagemEntrada;
    int *imagemSaida;
    Prototipo *mediaPrototipos;

    printf("\nAnalise de Textura por Histograma de somas e diferencas <Anderson Rocha>\n");

    if (argc == 5)
    {
        // Imagem de entrada a ser classificada
        OK = le_imagem_pgm (argv[1], &imagemEntrada, &nl, &nc, &mn);
        // Maior nivel de cinza da imagem de entrada
        MAX_NIVEL_CINZA = mn;
        // Alocacao de memoria para a imagem de resposta
        aloca_memo (&imagemSaida, nl, nc);
        // Imprimindo informacoes para o usuario
        info_imagem (argv[1], nl, nc, mn);
        // Lendo os parametros
        configurarParametros(argv[3]);

        printf("TAM_JANELA = %d, DIRX = %d, DIRY = %d\n", TAM_JANELA, DIRX, DIRY);
        printf("NRO_PROTOTIPOS = %d, NRO_DESCRITORES = %d\n\n", NRO_PROTOTIPOS, NRO_DESCRITORES);

        // Criando o vetor de prototipos
        mediaPrototipos = calloc(NRO_PROTOTIPOS, sizeof(Prototipo));

        // Inicializando os descritores para cada prototipo
        for(i = 0; i < NRO_PROTOTIPOS; i++)
        {
            mediaPrototipos[i].descritor = calloc(NRO_DESCRITORES, sizeof(double));

            for(j = 0; j < NRO_DESCRITORES; j++)
                mediaPrototipos[i].descritor[j] = 0.0;
        }

        // Caso 0 = Fase de apresendizado e gravacao em um arquivo
        // Caso 1 = Fase de classificacao. Parametros sao lidos do arquivo
        if(atoi(argv[4]) == 0)
            configurarPrototipos(ARQUIVO_PROTOTIPOS, mediaPrototipos);
        else
            classificar(imagemEntrada, imagemSaida, nl, nc, mediaPrototipos);

        if(imagemEntrada != null)
        {
            grava_imagem_pgm (imagemSaida, argv[2], nl, nc, mn);
            printf("Imagem \"%s\" gravada com sucesso!\n", argv[2]);
        }
        else
        {
            printf("Nao foi possivel gravar o resultado\n");
        }

        desaloca_memo(&imagemSaida);
    }
    else
    {
        printf("Sintaxe: atex <img_entrada> <img_saida> <arquivo_configuracao> <aprendizado = 0, classificacao = 1> \n");
    }

    // Desalocando as variaveis
    for(i = 0; i < NRO_PROTOTIPOS; i++)
        free(mediaPrototipos[i].descritor);

    free(mediaPrototipos);

    printf("\n");

    return(0);
}

// Lendo o arquivo de prototipos criado na fase de aprendizado
void lerPrototipos(Prototipo *p)
{
    int i, j;
    char str[40];
    FILE *fp;

    i = 0;
    j = 0;

    /* Arquivo ASCII, para leitura */
    fp = fopen("prototipos.cfg","r");

    if(!fp)
    {
        printf("Erro na abertura do arquivo");
        exit(0);
    }

    while(fgets(str, 40, fp))
    {
        p[i].descritor[j] = atof(str);
        j = (j + 1) % NRO_DESCRITORES;

        if(j == 0)
            i = (i + 1) % NRO_PROTOTIPOS;
    }
}

// Classificacao da imagem de entrada a partir dos prototipos fornecidos
void classificar(int *imagem, int *imagemSaida, int nl, int nc, Prototipo *p)
{
        double d[NRO_PROTOTIPOS];
        double menorDistancia;
        int posMenor;

        // Uma imagem para cada descritor
        int *vetorImagens[NRO_DESCRITORES];
        int *img;
        int *tmp;
        int i, j, k, w, z;
        // Conjunto de descritores da janela sendo analisada
        Prototipo descAlvo;
        // Conjunto de descritores mdios do prototipo em questao
        Prototipo descProt;

        // Criando o descritor alvo
        descAlvo.descritor = calloc(NRO_DESCRITORES, sizeof(double));
        // Criando o descritor comparativo
        descProt.descritor = calloc(NRO_DESCRITORES, sizeof(double));

        // Configurando descritores a partir do arquivo
        lerPrototipos(p);

        // Impressao dos prototipos sendo utilizados
        printf("Descritores para comparacao\n");
        for(i = 0; i < NRO_PROTOTIPOS; i++)
        {
            printf("[");
            for(j = 0; j < NRO_DESCRITORES; j++)
            {
                printf("%f\t", p[i].descritor[j]);
            }
            printf("]\n");
        }
        printf("\n");

        // Alocando uma imagem para cada descritor possivel
        for(i = 0; i < NRO_DESCRITORES; i++)
        {
            aloca_memo (&img, nl, nc);
            vetorImagens[i] = img;
        }

        for(i = 0; i < NRO_DESCRITORES; i++)
        {
            detectarTexturas(imagem, vetorImagens[i], nl, nc);
            DESCRITOR++;
        }

        // Calculando as distancias euclidianas
        for(i = 0; i < NRO_DESCRITORES; i++)
            descAlvo.descritor[i] = 0.0;

        for(i = 0; i < nl; i++)
        {
            for(j = 0; j < nc; j++)
            {
                // Cada pixel  um descritor que deve ser associado ao descAlvo
                for(w = 0; w < NRO_DESCRITORES; w++)
                {
                    tmp = vetorImagens[w];
                    descAlvo.descritor[w] = *(tmp + i * nc + j);
                }

                // Calculando a distancia do pixel central a cada prototipo
                for(w = 0; w < NRO_PROTOTIPOS; w++)
                {
                    descProt = p[w];
                    d[w] = distanciaEuclidiana(&descAlvo, &descProt, NRO_DESCRITORES);
                }

                // Achando a menor distancia euclidiana calculada
                menorDistancia = d[0];
                posMenor = 0;
                for(w = 1; w < NRO_PROTOTIPOS; w++)
                {
                    if(d[w] < menorDistancia)
                    {
                        menorDistancia = d[w];
                        posMenor = w;
                    }
                }

                // Selecionando a cor de saida
                *(imagemSaida + i * nc + j) = (MAX_NIVEL_CINZA / NRO_PROTOTIPOS) * posMenor;
            }
        }

        // Liberando espaco de memoria
        free (descProt.descritor);
        free (descAlvo.descritor);
}

// Calculo da distancia Euclidiana. DE = sqrt[(x1 - x2)^2 - (y1 - y2)^2]
double distanciaEuclidiana(Prototipo *origem, Prototipo *destino, int n)
{
    double distancia;
    int i;

    distancia = 0.0;

    for(i = 0; i < n; i++)
        distancia += pow((origem->descritor[i] - destino->descritor[i]), 2);

    return (sqrt(distancia));
}

void configurarParametros(char *nomeArquivo)
{
    FILE *fp;
    char str[40];

    /* Arquivo ASCII, para leitura */
    fp = fopen(nomeArquivo,"r");

    if(!fp)
    {
        printf("Erro na abertura do arquivo");
        exit(0);
    }

    fgets(str, 40, fp);
    TAM_JANELA = atoi(str);
    fgets(str, 40, fp);
    DIRX = atoi(str);
    fgets(str, 40, fp);
    DIRY = atoi(str);
    fgets(str, 40, fp);
    NRO_PROTOTIPOS = atoi(str);
    fgets(str, 40, fp);
    NRO_DESCRITORES = atoi(str);
    fgets(str, 40, fp);
    TODAS_DIRECOES = atoi(str);

    fgets(ARQUIVO_PROTOTIPOS, 40, fp);
    DESCRITOR = 0;
}

// Criando os prototipos a partir dos nomes de prototipos fornecidos como parametro
void configurarPrototipos(char *nomeArquivo, Prototipo *p)
{
    FILE *fp, *fp1;
    int OK, i, j, k, contador, nc, nl, mn, tmp;
    int *imagem;
    char str[40];
    char *prototipos[NRO_PROTOTIPOS];
    char *nome;
    // Calcula um descritor em todo o prototipo
    int *umPrototipo;
    // Para um descritor quantas vezes ele foi diferente em uma imagem prototipo
    int nroOcorrenciasDescritor;
    // Como sao varios valores diferentes precisamos saber a soma de todos os valores
    // para depois calcular a media que define este descritor
    float somaDescritor, media;

    contador = 0;
    somaDescritor = 0.0;
    media = 0.0;

    /* Arquivo ASCII, para leitura */
    nome =  calloc (strlen(nomeArquivo) - 1, sizeof(char));
    strncpy(nome, nomeArquivo, strlen(nomeArquivo) - 1);
    fp = fopen(nome,"r");
    // Criando o arquivo de prototipos
    fp1 = fopen("prototipos.cfg","w");

    free(nome);

    if(!fp)
    {
        printf("Erro na abertura do arquivo");
        exit(0);
    }

    while((contador < NRO_PROTOTIPOS) && (fgets(str, 40, fp) != NULL))
    {
        nome = calloc (strlen(str) - 1, sizeof(char));
        strncpy(nome, str, strlen(str) - 1);
        prototipos[contador] = nome;
        contador++;
    }

    // Assegurando que o nroPrototipos seja realmente o numero de prototipos que
    // foi lido do arquivo de configuracao
    NRO_PROTOTIPOS = contador;

    printf("\nPrototipos\n");
    printf("----------\n");
    for(i = 0; i < NRO_PROTOTIPOS; i++)
    {
        printf("%s\n", prototipos[i]);
        OK = le_imagem_pgm (prototipos[i], &imagem, &nl, &nc, &mn);

        for(DESCRITOR = 0; DESCRITOR < NRO_DESCRITORES; DESCRITOR++)
        {
            aloca_memo (&umPrototipo, nl, nc);
            detectarTexturas(imagem, umPrototipo, nl, nc);
            nroOcorrenciasDescritor = (nl - TAM_JANELA + 1) * (nc - TAM_JANELA + 1);
            somaDescritor = 0;

            // Calculando a media do descritor
            for(j = 0; j < nl; j++)
            {
                for(k = 0; k < nc; k++)
                {
                    tmp = *(umPrototipo + j * nc + k);

                    if(tmp != 0);
                        somaDescritor += tmp;
                }
            }

            // Gravacao em arquivo do prototipo calculado
            sprintf(str, "%f\n", (float)(somaDescritor/(float)nroOcorrenciasDescritor));
            fputs(str, fp1);

            // Impressao na tela. Feedback para o usuario
            printf("Media DESCRITOR %d = %f\n",
                    DESCRITOR, (float)(somaDescritor/(float)nroOcorrenciasDescritor));
            p[i].descritor[DESCRITOR] = (float)(somaDescritor/(float)nroOcorrenciasDescritor);
        }
    }

    printf("\n");

    for(i = 0; i < NRO_PROTOTIPOS; i++)
        free(prototipos[i]);

    fclose(fp);
    fclose(fp1);

    // Deixando DESCRITOR media como padro
    DESCRITOR = 0;


}


// Imagem[i,j] = *(Imagem + i * nc + j)
// Caso o usuario queira trabalhar com histogramas menores
void subQuantizarImagem(int *imagem, int nl, int nc,  int maxCinza)
{
    int i, j;

    for(i = 0; i < nl; i++)
    {
        for(j = 0; j < nc; j++)
            *(imagem + i * nc + j) = *(imagem + i * nc + j) / maxCinza;
    }
}

// Para um dado descritor, calcula o valor do elemento central da janela
// para cada janela possivel. Retorna uma imagem com seus pontos centrais
// alterados
void detectarTexturas(int *imagem, int *imagemSaida, int nl, int nc)
{
    // coluna,  linha, numero de janelas, nro colunas da janela, nro linhas
    int c, l, nj, i, j, k, indice, i1;
    int janela[TAM_JANELA][TAM_JANELA];
    int x;
    int elementoCentral;
    int contador;
    // Reflete qual o elemento da janela passada deve ser alterado com o
    // valor de k passado
    int posicao;
    int hs[2*MAX_NIVEL_CINZA];
    int hd[2*MAX_NIVEL_CINZA];
    double valorDescritor;

    double vetorDescritores[8];

    contador = 0;
    posicao = (TAM_JANELA * TAM_JANELA)/ 2 + 1;
    elementoCentral = 0;
    nj = 0;
    x = 0;

    // Inicializando imagemSaida como zero
    for(i = 0; i < nl; i++)
    {
        for(j = 0; j < nc; j++)
        {
            *(imagemSaida + i * nc + j) = 0;
        }
    }

    for(x = 0; x < nl - TAM_JANELA + 1; x++)
    {
        for(nj = 0; nj < nc - TAM_JANELA + 1; nj++)
        {
            k = nj + x * nc;
            l = 0;
            contador = 0;

            for(i = 0; i < TAM_JANELA; i++)
            {
                c = 0;
                for(j = k; j < (TAM_JANELA + k); j++)
                {
                    contador++;
                    janela[l][c] = *(imagem + j);
                    c++;

                    if(contador == posicao)
                        elementoCentral = j;

                }
                k = (k + nc);
                l++;
            }

            valorDescritor = 0.0;

            // Faz apenas para a direcao passada como parametro
            if(TODAS_DIRECOES == 0)
            {
                criarHistogramaSD(janela, TAM_JANELA, TAM_JANELA, DIRX, DIRY, hs, hd);
                valorDescritor = avaliarDescritor(hs, hd, DESCRITOR);
            }
            else
            {
                // Computo em todas as direes possveis. Explorando ao mximo
                // a vizinhana
                criarHistogramaSD(janela, TAM_JANELA, TAM_JANELA, 0, 1, hs, hd);
                vetorDescritores[0] = avaliarDescritor(hs, hd, DESCRITOR);
                criarHistogramaSD(janela, TAM_JANELA, TAM_JANELA, 1, 0, hs, hd);
                vetorDescritores[1] = avaliarDescritor(hs, hd, DESCRITOR);
                criarHistogramaSD(janela, TAM_JANELA, TAM_JANELA, 1, 1, hs, hd);
                vetorDescritores[2] = avaliarDescritor(hs, hd, DESCRITOR);
                criarHistogramaSD(janela, TAM_JANELA, TAM_JANELA, -1, 0, hs, hd);
                vetorDescritores[3] = avaliarDescritor(hs, hd, DESCRITOR);
                criarHistogramaSD(janela, TAM_JANELA, TAM_JANELA, 0, -1, hs, hd);
                vetorDescritores[4] = avaliarDescritor(hs, hd, DESCRITOR);
                criarHistogramaSD(janela, TAM_JANELA, TAM_JANELA, -1,-1, hs, hd);
                vetorDescritores[5] = avaliarDescritor(hs, hd, DESCRITOR);
                criarHistogramaSD(janela, TAM_JANELA, TAM_JANELA, 1, -1, hs, hd);
                vetorDescritores[6] = avaliarDescritor(hs, hd, DESCRITOR);
                criarHistogramaSD(janela, TAM_JANELA, TAM_JANELA, -1, 1, hs, hd);
                vetorDescritores[7] = avaliarDescritor(hs, hd, DESCRITOR);

                for(i1 = 0; i1 < 8; i1++)
                    valorDescritor += vetorDescritores[i1];

                valorDescritor = valorDescritor / 8;
            }

            *(imagemSaida + elementoCentral) = (int)valorDescritor;
       }
    }
}

// Avalia um dado descritor
double avaliarDescritor(int *hs, int *hd, int descritor)
{
    int i, j, nroElementos, nroElementos2;
    double contraste, fator, media, homogen, energiaX, energiaY, varX, varY;
    double correlacaoX, correlacaoY;

    contraste = 0.0;
    fator = 0.0;
    media = 0.0;
    homogen = 0.0;
    energiaX = 0.0;
    energiaY = 0.0;
    varX = 0.0;
    varY = 0.0;
    correlacaoX = 0.0;
    correlacaoY = 0.0;
    nroElementos = 0;

    switch(descritor)
    {
        case 0:
            // Descritor media
            for(i = 0; i < 2 * MAX_NIVEL_CINZA; i++)
            {
                if(hs[i] != 0)
                {
                    media += (double)(i * hs[i]);
                    nroElementos += hs[i];
                }
            }

            media = media / (double)(2 * nroElementos);

            return (double)(media);
        case 1:
            // Descritor Contraste
            contraste = 0.0;
            nroElementos = 0;

            for(i = 0; i < 2 * MAX_NIVEL_CINZA; i++)
            {
                if(hd[i] != 0)
                {
                    contraste += (pow((i - MAX_NIVEL_CINZA),2)) * (hd[i]);
                    nroElementos +=  hd[i];
                }
            }

            return (double)(contraste / nroElementos);
        case 2:
            // Descritor de homogeneidade
            for(i = 0; i < 2 * MAX_NIVEL_CINZA; i++)
            {
                if(hd[i] != 0)
                {
                    fator = (double) (1.0 / (1.0 + pow((double)(i - MAX_NIVEL_CINZA), 2)));
                    homogen += (double)(hd[i] * fator);
                    nroElementos += hd[i];
                }
            }

            return homogen;
        case 3:
            // Descritor energia

            for(i = 0; i < 2 * MAX_NIVEL_CINZA; i++)
            {
                energiaX += pow(hs[i],2);
                energiaY += pow(hd[i],2);
                nroElementos += hs[i];
            }

            return (double)(energiaX * energiaY) / nroElementos;
        case 4:
            // Descritor variancia
            media = avaliarDescritor(hs, hd, 0);

            for(i = 0; i < 2 * MAX_NIVEL_CINZA; i++)
            {
                varX += pow((i - 2 * media),2) * hs[i];
                varY += pow((i - MAX_NIVEL_CINZA),2) * hd[i];
                nroElementos += hs[i];
            }

            return ((varX + varY)/ (nroElementos));
        case 5:
            // Correlacao
            media = avaliarDescritor(hs, hd, 0);

            for(i = 0; i < 2 * MAX_NIVEL_CINZA; i++)
            {
                correlacaoX += pow((i - 2 * media),2) * hs[i];
                correlacaoY += pow((i - MAX_NIVEL_CINZA),2) * hd[i];
                nroElementos += hs[i];
            }

            return (0.5 * (correlacaoX - correlacaoY) / (nroElementos));
        case 6:
            // Entropia

            for(i = 0; i < 2 * MAX_NIVEL_CINZA; i++)
            {
                if(hs[i] != 0)
                    varX += (log(hs[i]) * hs[i]);
                if(hd[i] != 0)
                    varY += (log(hd[i]) * hd[i]);
            }

            return (-1) * (varX + varY);
        default:
            return null;
    }
}

/*
    imagem = imagem armazenada
    nl = numero de linhas
    nc = numero de colunas
    h = histograma
    Imagem[i,j] = *(Imagem + i * nc + j)
    Cria o histograma das somas e das diferencas para analise posterior
*/
void criarHistogramaSD(int janela[][TAM_JANELA], int nl, int nc, int t1, int t2, int *hs, int *hd)
{
    int offsetX, offsetY, i, j, k;
    // Contador de ocorrencias
    int ocorrencias = 0;

    // Zerando os histogramas
    for(i = 0; i < 2 * MAX_NIVEL_CINZA; i++)
    {
        hs[i] = 0;
        hd[i] = 0;
    }

    // Histograma de somas
    for(i = 0; i < nl; i++)
    {
        for(j = 0; j < nc; j++)
        {
            offsetX = i + t1;
            offsetY = j + t2;

            if((offsetX < nl) && (offsetX > 0) && (offsetY < nc) && (offsetY > 0))
            {
                hs[janela[i][j] + janela[i + t1][j + t2]] += 1;
                hd[janela[i][j] - janela[i+t1][j + t2] + MAX_NIVEL_CINZA] += 1;
                ocorrencias++;
            }
        }
    }
}
