/*
    Trabalho Prtico III - IPDI
    -------------------------
    Anderson de Rezende Rocha <anderson.rocha@ic.unicamp.br>
    Anlise de texturas por comparao de histogramas

    OBSERVAO
    ----------
    Imagem[i,j] = *(Imagem + i * nc + j)
    
    Referncias http://www.bmva.ac.uk/bmvc/1997/papers/066/node4.html
    Minkowsky http://disney.ctr.columbia.edu/jrsthesis/node66.html
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "imagem.h"
#include "morfo.h"
#include "utils.h"

// Constante para alguma referncia de nmeros muito grandes
#define INFINITO 99999999
// Referncia ao ponteiro /dev/null
#define null 0
// Histograma de somas
#define HIST_SOMA 0
// Histograma de diferenas
#define HIST_DIF 1

typedef struct T_Prototipo
{
    int *histograma;
} Prototipo;

static int TAM_JANELA;
static int MAX_NIVEL_CINZA;
static int NRO_PROTOTIPOS;
static int NRO_DESCRITORES = 3;
static int METRICA;
static char ARQUIVO_PROTOTIPOS[40];

void configurarParametros(char *nomeArquivo);
void detectarTexturas(int *imagem, int *imagemSaida, int nl, int nc, Prototipo *p);
void imprimirHistograma(int *hist);
void configurarPrototipos(char *nomeArquivo, Prototipo *p);
void criarHistograma(int *janela, int nl, int nc, int *hist);
double calcularMetrica(int *h1, int *h2, int metrica);
double modulo(double i);

int main (int argc, char *argv[])
{
    int OK, nc, nl, mn, i, j;
    int *imagemEntrada;
    int *imagemSaida;
    Prototipo *prototipos;

    printf("\nAnalise de Textura por comparacao de histogramas <Anderson Rocha>\n");

    if (argc == 4)
    {
        OK = le_imagem_pgm (argv[1], &imagemEntrada, &nl, &nc, &mn);
        MAX_NIVEL_CINZA = mn;
        aloca_memo (&imagemSaida, nl, nc);
        info_imagem (argv[1], nl, nc, mn);
        configurarParametros(argv[3]);

        printf("TAM_JANELA = %d, NRO_PROTOTIPOS = %d\n", TAM_JANELA, NRO_PROTOTIPOS);
        printf("Arquivo prototipos = %s\n", ARQUIVO_PROTOTIPOS);

        // Criando o vetor de prototipos
        prototipos = calloc(NRO_PROTOTIPOS, sizeof(Prototipo));

        // Inicializando os descritores para cada prototipo
        for(i = 0; i < NRO_PROTOTIPOS; i++)
        {
            prototipos[i].histograma = calloc(MAX_NIVEL_CINZA, sizeof(int));

            for(j = 0; j < MAX_NIVEL_CINZA; j++)
                prototipos[i].histograma[j] = 0;
        }

        configurarPrototipos(ARQUIVO_PROTOTIPOS, prototipos);
        detectarTexturas(imagemEntrada, imagemSaida, nl, nc, prototipos);

        if(imagemEntrada != null)
        {
            grava_imagem_pgm (imagemSaida, argv[2], nl, nc, mn);
            printf("Imagem \"%s\" gravada com sucesso!\n", argv[2]);
        }
        else
        {
            printf("No foi possivel gravar o resultado\n");
        }

        desaloca_memo(&imagemSaida);
    }
    else
    {
        printf("Erro de sintaxe\n");
    }

    printf("\n");

    // Desalocando as variaveis
    for(i = 0; i < NRO_PROTOTIPOS; i++)
        free(prototipos[i].histograma);

    free(prototipos);

    return(0);
}

void configurarParametros(char *nomeArquivo)
{
    FILE *fp;
    char str[40];

    /* Arquivo ASCII, para leitura */
    fp = fopen(nomeArquivo,"r");

    if(!fp)
    {
        printf("Erro na abertura do arquivo");
        exit(0);
    }

    fgets(str, 40, fp);
    TAM_JANELA = atoi(str);
    fgets(str, 40, fp);
    METRICA = atoi(str);
    fgets(str, 40, fp);
    NRO_PROTOTIPOS = atoi(str);
    fgets(str, 40, fp);
    strncpy(ARQUIVO_PROTOTIPOS, str, strlen(str) - 1);
}

void configurarPrototipos(char *nomeArquivo, Prototipo *p)
{
    FILE *fp, *fp1;
    int OK, i, j, k, contador, nc, nl, mn, tmp;
    int *imagem;
    char str[40];
    char *prototipos[NRO_PROTOTIPOS];
    char *nome;

    contador = 0;

    /* Arquivo ASCII, para leitura */
    fp = fopen(nomeArquivo,"r");

    // Criando o arquivo de prototipos para o aprendizado
    fp1 = fopen("prototipos.cfg","w");

    if(!fp)
    {
        printf("Erro na abertura do arquivo");
        exit(0);
    }

    while((contador < NRO_PROTOTIPOS) && (fgets(str, 40, fp) != NULL))
    {
        nome = calloc (strlen(str) - 1, sizeof(char));
        strncpy(nome, str, strlen(str) - 1);
        prototipos[contador] = nome;
        contador++;
    }

    // Assegurando que o nroPrototipos seja realmente o numero de prototipos que
    // foi lido do arquivo de configuracao
    NRO_PROTOTIPOS = contador;

    printf("\nPrototipos\n");
    printf("----------\n");
    for(i = 0; i < NRO_PROTOTIPOS; i++)
    {
        printf("%s\n", prototipos[i]);
        OK = le_imagem_pgm (prototipos[i], &imagem, &nl, &nc, &mn);

        criarHistograma(imagem, nl, nc, p[i].histograma);
    }

    printf("\n");

    for(i = 0; i < NRO_PROTOTIPOS; i++)
        free(prototipos[i]);

    fclose(fp);
    fclose(fp1);
}

// Calcula a metrica indicada para a relacao entre os
// histogramas H1 e H2 passados
double calcularMetrica(int *h1, int *h2, int metrica)
{
    int i, j, nroElementosH1, nroElementosH2;
    double soma, fator, somaH1H2;
    // Os novos histogramas normalizados
    double h1N[MAX_NIVEL_CINZA], h2N[MAX_NIVEL_CINZA];

    soma = 0.0;
    fator = 0.0;
    somaH1H2 = 0.0;
    nroElementosH1 = 0;
    nroElementosH2 = 0;

    // Normalizando os dois histogramas para pode comparar
    for(i = 0; i < MAX_NIVEL_CINZA; i++)
    {
        nroElementosH1 += h1[i];
        nroElementosH2 += h2[i];
        h1N[i] = 0.0;
        h2N[i] = 0.0;
    }

    for(i = 0; i < MAX_NIVEL_CINZA; i++)
    {
        h1N[i] = (double)((double)h1[i] / (double)nroElementosH1);
        h2N[i] = (double)((double)h2[i] / (double)nroElementosH2);
    }

    // A Norma L1 e L2 so baseadas na distncia de Minkowsky onde M = Soma[i=0][M](|h1(i) - h2(i)|^f)
    // onde f  uma funo a ser definida
    switch(metrica)
    {
        case 0:
            // Norma L1
            for(i = 0; i < MAX_NIVEL_CINZA; i++)
                soma += modulo(h1N[i] - h2N[i]);

            return soma;
            break;
        case 1:
            // Norma L2
            for(i = 0; i < MAX_NIVEL_CINZA; i++)
                soma += pow((h1N[i] - h2N[i]), 2);

            return sqrt(soma);
            break;
        case 2:
            // Chi-Square
            for(i = 0; i < MAX_NIVEL_CINZA; i++)
            {
                fator = pow((h1N[i] - h2N[i]), 2);
                somaH1H2 = (h1N[i] + h2N[i]);

                if(somaH1H2 > 0.0)
                {
                    soma += (double)(fator / somaH1H2);
                }
            }

            return soma;
            break;
        case 3:
            // Bhattacharyya
            for(i = 0; i < MAX_NIVEL_CINZA; i++)
            {
                fator = sqrt(h1N[i] * h2N[i]);
                soma += fator;
            }

            return (log(soma) / log(2.71828183));
            break;
        default:
            printf("Metrica nao definida\n");
            break;
    }

    return null;
}

double modulo(double i)
{
    return (i > 0 ? i : (-1.0) * i);
}

void detectarTexturas(int *imagem, int *imagemSaida, int nl, int nc, Prototipo *p)
{
    // coluna,  linha, numero de janelas, nro colunas da janela, nro linhas
    int c, l, nj, i, j, k, indice, i1;
    int *janela;
    int x;
    int elementoCentral;
    int contador;
    // Reflete qual o elemento da janela passada deve ser alterado com o
    // valor de k passado
    int posicao;
    int hist[MAX_NIVEL_CINZA];
    double metricas[NRO_PROTOTIPOS];
    int menorMetrica;

    contador = 0;
    posicao = (TAM_JANELA * TAM_JANELA)/ 2 + 1;
    elementoCentral = 0;
    nj = 0;
    x = 0;

    // Alocando espaco para a janela que iremos utilizar
    aloca_memo (&janela, TAM_JANELA, TAM_JANELA);

    // Inicializando imagemSaida como zero
    for(i = 0; i < nl; i++)
    {
        for(j = 0; j < nc; j++)
        {
            *(imagemSaida + i * nc + j) = 0;
        }
    }

    for(x = 0; x < nl - TAM_JANELA + 1; x++)
    {
        for(nj = 0; nj < nc - TAM_JANELA + 1; nj++)
        {
            k = nj + x * nc;
            l = 0;
            contador = 0;

            for(i = 0; i < TAM_JANELA; i++)
            {
                c = 0;
                for(j = k; j < (TAM_JANELA + k); j++)
                {
                    contador++;
                    *(janela + l * TAM_JANELA + c) = *(imagem + j);
                    c++;

                    if(contador == posicao)
                        elementoCentral = j;

                }
                k = (k + nc);
                l++;
            }

            criarHistograma(janela, TAM_JANELA, TAM_JANELA, hist);

            for(i1 = 0; i1 < NRO_PROTOTIPOS; i1++)
            {
                metricas[i1] = calcularMetrica(hist, p[i1].histograma, METRICA);
            }

            menorMetrica = 0;

            for(i1 = 1; i1 < NRO_PROTOTIPOS; i1++)
            {
                if(metricas[i1] <= metricas[menorMetrica])
                    menorMetrica = i1;
            }

            *(imagemSaida + elementoCentral) = (MAX_NIVEL_CINZA / NRO_PROTOTIPOS) * menorMetrica;
       }
    }

    desaloca_memo(&janela);
}

void imprimirHistograma(int *hist)
{
    int i;

    for(i = 0; i < MAX_NIVEL_CINZA; i++)
        printf("h[%i] = %d\n", i, *(hist + i));

    printf("\n");
}

/*
    imagem = imagem armazenada
    nl = numero de linhas
    nc = numero de colunas
    h = histograma
    Imagem[i,j] = *(Imagem + i * nc + j)
*/
void criarHistograma(int *janela, int nl, int nc, int *hist)
{
    int i, j;

    // Zerando os histogramas
    for(i = 0; i < MAX_NIVEL_CINZA; i++)
        hist[i] = 0;

    // Histograma de somas
    for(i = 0; i < nl; i++)
    {
        for(j = 0; j < nc; j++)
        {
             hist[*(janela + i * nc + j)] += 1;
        }
    }
}
